import torch

class Metric:
    """ Base class for all metrics"""
    def __init__(self, name: str) -> None:
        """ Initialize metric with name

        Args:
            name (str): name of metric
        """
        self.name = name

    def reset(self):
        """ Reset metric state to initial values and return metric value"""
        self.__init__()

    def update(self, output: torch.Tensor, target: torch.Tensor):
        """ Update metric state with new data
        
        Args:
            output (torch.Tensor): output of model
            target (torch.Tensor): target of data
        """
        pass

    def result(self):
        """ Return metric value"""
        pass


class Accuracy(Metric):
    """ Accuracy metric class
    
    Args:
        name (str, optional): name of metric. Defaults to 'accuracy'.
    """
    def __init__(self, name='accuracy') -> None:
        super(Accuracy, self).__init__(name=name)
        self.correct = 0
        self.total = 0

    def update(self, output: torch.Tensor, target: torch.Tensor):
        """ Update metric state with new data

        Args:
            output (torch.Tensor): output of model
            target (torch.Tensor): target of data
        """
        _, predicted = torch.max(output.data, 1)
        self.total += target.size(0)
        self.correct += (predicted == target).sum().item()

    def result(self):
        """ Return metric value"""
        return self.correct / self.total


import cv2
import typing
import numpy as np

""" Implemented augmentors:
- RandomBrightness
- RandomRotate
- RandomErodeDilate
- RandomSharpen
"""

class Augmentor:
    """ Object that should be inherited by all augmentors

    Args:
        random_chance (float, optional): Chance of applying the augmentor. Where 0.0 is never and 1.0 is always. Defaults to 0.5.
    """
    def __init__(self, random_chance: float=0.5) -> None:
        self._random_chance = random_chance

    def __call__(self, image: np.ndarray, annotation: np.ndarray) -> typing.Tuple[np.ndarray, np.ndarray]:
        if np.random.random() <= self._random_chance:
            pass

        return image, annotation

class RandomBrightness(Augmentor):
    """ Randomly adjust image brightness

    Attributes:
        random_chance (float): Float between 0.0 and 1.0 setting bounds for random probability. Defaults to 0.5.
        delta (int): Integer value for brightness adjustment
    """
    def __init__(
        self, 
        delta: int = 100,
        *args, **kwargs
        ) -> None:
        super(RandomBrightness, self).__init__(*args, **kwargs)

        assert 0 <= delta <= 255.0, "Delta must be between 0.0 and 255.0"

        self._delta = delta

    def __call__(self, image: np.ndarray, annotation: np.ndarray) -> typing.Tuple[np.ndarray, np.ndarray]:
        """ Randomly adjust image brightness

        Args:
            image (np.ndarray): Image to be adjusted
            annotation (np.ndarray): Mask to be adjusted

        Returns:
            image (np.ndarray): Adjusted image
            annotation (np.ndarray): Adjusted mask
        """
        if np.random.rand() <= self._random_chance:

            image = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)

            value = 1 + np.random.uniform(-self._delta, self._delta) / 255

            hsv = np.array(image, dtype = np.float32)

            hsv[:, :, 1] = hsv[:, :, 1] * value
            hsv[:, :, 2] = hsv[:, :, 2] * value

            hsv = np.uint8(np.clip(hsv, 0, 255))

            image = cv2.cvtColor(hsv, cv2.COLOR_HSV2BGR)

        return image, annotation

class RandomRotate(Augmentor):
    """ Randomly rotate image

    Attributes:
        random_chance (float): Float between 0.0 and 1.0 setting bounds for random probability. Defaults to 0.5.
        angle (int): Angle between 0 and 180, setting image rotation borders
        borderValue (tuple): Tuple of 3 integers, setting border color for image rotation
        crop_borders (bool): Boolean value, setting if borders should be cropped after rotation
    """
    def __init__(
        self, 
        angle: int=30, 
        borderValue: typing.Tuple[int, int, int]=None,
        crop_borders: bool=False,
        *args, **kwargs
        ) -> None:
        super(RandomRotate, self).__init__(*args, **kwargs)

        self._angle = angle
        self._borderValue = borderValue
        self._crop_borders = crop_borders

    def __call__(self, image:np.ndarray, annotation:np.ndarray)->typing.Tuple[np.ndarray, np.ndarray]:
        """ Randomly rotate image

        Args:
            image (np.ndarray): Image to be rotated
            annotation (np.ndarray): Mask to be rotated

        Returns:
            image (np.ndarray): Rotated image
            annotation (np.ndarray): Rotated mask
        """
        if np.random.rand() <= self._random_chance:

            angle = np.random.uniform(-self._angle, self._angle)

            # generate random border color
            borderValue = np.random.randint(0, 255, 3) if self._borderValue is None else self._borderValue
            borderValue = [int(v) for v in borderValue]

            h, w, _ = image.shape
            if self._crop_borders:
                m = cv2.getRotationMatrix2D((w / 2, h / 2), angle, 1)
                image = cv2.warpAffine(image, m, (w, h), borderValue=borderValue)
                # Check if annotation is image mask
                if not isinstance(annotation, str):
                    annotation = cv2.warpAffine(annotation, m, (w, h), borderValue=0)

            else:                
                diagonal = round(np.sqrt((w*w) + (h*h)))
                top, bottom, left, right = round((diagonal-h) / 2), round((diagonal-h) / 2), round((diagonal-w) / 2), round((diagonal-w) / 2)
                padded_image = cv2.copyMakeBorder(image, top, bottom, left, right, borderType=cv2.BORDER_CONSTANT, value = borderValue)

                padded_height, padded_width = padded_image.shape[:2]

                transform_matrix = cv2.getRotationMatrix2D((padded_height/2, padded_width/2), angle, 1.0)

                rotated_image = cv2.warpAffine(padded_image, transform_matrix, (diagonal, diagonal), flags=cv2.INTER_LANCZOS4, borderValue=borderValue)

                # Find the indices of the non-black pixels in the image
                indices = np.argwhere(rotated_image != np.array(borderValue))

                # Find the minimum and maximum row and column indices of the non-black pixels
                min_row, max_row = indices[:, 0].min(), indices[:, 0].max()
                min_col, max_col = indices[:, 1].min(), indices[:, 1].max()

                # Crop the black borders from the image
                cropped_image = rotated_image[min_row:max_row+1, min_col:max_col+1]

                if not isinstance(annotation, str):

                    padded_annotation = cv2.copyMakeBorder(annotation, top, bottom, left, right, borderType=cv2.BORDER_CONSTANT, value=0)

                    rotated_annotation = cv2.warpAffine(padded_annotation, transform_matrix, (diagonal, diagonal), flags=cv2.INTER_LANCZOS4, borderValue=0)

                    cropped_annotation = rotated_annotation[min_row:max_row+1, min_col:max_col+1]

                    return cropped_image, cropped_annotation

                return cropped_image, annotation

        return image, annotation

class RandomErodeDilate(Augmentor):
    """ Randomly erode and dilate image

    Attributes:
        random_chance (float): Float between 0.0 and 1.0 setting bounds for random probability. Defaults to 0.5.
        kernel_size (tuple): Tuple of 2 integers, setting kernel size for erosion and dilation
    """
    def __init__(
        self, 
        kernel_size: typing.Tuple[int, int]=(1, 1), 
        *args, **kwargs
        ) -> None:
        super(RandomErodeDilate, self).__init__(*args, **kwargs)

        self._kernel_size = kernel_size

    def __call__(self, image:np.ndarray, annotation)->typing.Tuple[np.ndarray, np.ndarray]:
        """ Randomly erode and dilate image

        Args:
            image (np.ndarray): Image to be eroded and dilated

        Returns:
            image (np.ndarray): Eroded and dilated image
        """
        if np.random.rand() <= self._random_chance:

            kernel = np.ones(self._kernel_size, np.uint8)

            if np.random.rand() <= 0.5:
                image = cv2.erode(image, kernel, iterations=1)
            else:
                image = cv2.dilate(image, kernel, iterations=1)

        return image, annotation

class RandomSharpen(Augmentor):
    """ Randomly sharpen image

    Attributes:
        alpha (float): Float between 0.0 and 1.0 setting bounds for random probability
        lightness_range (tuple): Tuple of 2 floats, setting bounds for random lightness change
        kernel (np.ndarray): Numpy array of kernel for image convolution
        kernel_anchor (np.ndarray): Numpy array of kernel anchor for image convolution
        random_chance (float, optional): Chance of applying the augmentor, where 1.0 is always and 0.0 is never. Defaults to 0.5.
    """
    def __init__(
        self, 
        alpha: float = 0.25,
        lightness_range: typing.Tuple = (0.75, 2.0),
        kernel: np.ndarray = None,
        kernel_anchor: np.ndarray = None,
        *args, **kwargs
        ) -> None:
        super(RandomSharpen, self).__init__(*args, **kwargs)

        self._alpha_range = (alpha, 1.0)
        self._ligtness_range = lightness_range
        self._lightness_anchor = 8

        self._kernel = np.array([[-1, -1, -1], [-1,  1, -1], [-1, -1, -1]], dtype=np.float32) if kernel is None else kernel
        self._kernel_anchor = np.array([[0, 0, 0], [0, 1, 0], [0, 0, 0]], dtype=np.float32) if kernel_anchor is None else kernel_anchor

        assert 0 <= alpha <= 1.0, "Alpha must be between 0.0 and 1.0"

    def __call__(self, image: np.ndarray, annotation) -> typing.Tuple[np.ndarray, np.ndarray]:
        """ Randomly sharpen image

        Args:
            image (np.ndarray): Image to be sharpened

        Returns:
            image (np.ndarray): Sharpened image
        """
        if np.random.rand() <= self._random_chance:

            lightness = np.random.uniform(*self._ligtness_range)
            alpha = np.random.uniform(*self._alpha_range)

            kernel = self._kernel_anchor  * (self._lightness_anchor + lightness) + self._kernel
            kernel -= self._kernel_anchor
            kernel = (1 - alpha) * self._kernel_anchor + alpha * kernel

            # Apply sharpening to each channel
            r, g, b = cv2.split(image)
            r_sharp = cv2.filter2D(r, -1, kernel)
            g_sharp = cv2.filter2D(g, -1, kernel)
            b_sharp = cv2.filter2D(b, -1, kernel)

            # Merge the sharpened channels back into the original image
            image = cv2.merge([r_sharp, g_sharp, b_sharp])

        return image, annotation


import cv2
import typing
import librosa
import librosa.display
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
matplotlib.interactive(False)

class ImageReader:
    """Read image with cv2 from path and return image and label"""
    def __init__(self, method: int = cv2.IMREAD_COLOR, *args, **kwargs):
        self._method = method

    def __call__(self, image_path: str, label: str):
        return cv2.imread(image_path, self._method), label

class WavReader:
    """Read wav file with librosa and return audio and label
    
    Attributes:
        frame_length (int): Length of the frames in samples.
        frame_step (int): Step size between frames in samples.
        fft_length (int): Number of FFT components.
    """
    def __init__(
        self, 
        frame_length: int=256, 
        frame_step: int=160,
        fft_length: int=384,
        *args, **kwargs
        ) -> None:
        self.frame_length = frame_length
        self.frame_step = frame_step
        self.fft_length = fft_length

    @staticmethod
    def get_spectrogram(wav_path: str, frame_length: int, frame_step: int, fft_length: int) -> np.ndarray:
        """Compute the spectrogram of a WAV file

        Args:
            wav_path (str): Path to the WAV file.
            frame_length (int): Length of the frames in samples.
            frame_step (int): Step size between frames in samples.
            fft_length (int): Number of FFT components.

        Returns:
            np.ndarray: Spectrogram of the WAV file.
        """
        # Load the wav file and store the audio data in the variable 'audio' and the sample rate in 'orig_sr'
        audio, orig_sr = librosa.load(wav_path) 

        # Compute the Short Time Fourier Transform (STFT) of the audio data and store it in the variable 'spectrogram'
        # The STFT is computed with a hop length of 'frame_step' samples, a window length of 'frame_length' samples, and 'fft_length' FFT components.
        # The resulting spectrogram is also transposed for convenience
        spectrogram = librosa.stft(audio, hop_length=frame_step, win_length=frame_length, n_fft=fft_length).T

        # Take the absolute value of the spectrogram to obtain the magnitude spectrum
        spectrogram = np.abs(spectrogram)

        # Take the square root of the magnitude spectrum to obtain the log spectrogram
        spectrogram = np.power(spectrogram, 0.5)

        # Normalize the spectrogram by subtracting the mean and dividing by the standard deviation.
        # A small value of 1e-10 is added to the denominator to prevent division by zero.
        spectrogram = (spectrogram - np.mean(spectrogram)) / (np.std(spectrogram) + 1e-10)

        return spectrogram

    @staticmethod
    def plot_raw_audio(wav_path: str, title:str = None, sr: int = 16000) -> None:
        """Plot the raw audio of a WAV file

        Args:
            wav_path (str): Path to the WAV file.
            sr (int, optional): Sample rate of the WAV file. Defaults to 16000.
        """
        # Load the wav file and store the audio data in the variable 'audio' and the sample rate in 'orig_sr'
        audio, orig_sr = librosa.load(wav_path, sr=sr)

        duration = len(audio) / orig_sr

        time = np.linspace(0, duration, num=len(audio))

        plt.figure(figsize=(15, 5))
        plt.plot(time, audio)
        plt.title(title) if title else plt.title('Audio Plot')
        plt.ylabel('signal wave')
        plt.xlabel('time (s)')
        plt.tight_layout()
        plt.show()

    @staticmethod
    def plot_spectrogram(spectrogram: np.ndarray, title:str = "", transpose: bool = True, invert: bool = True) -> None:
        """Plot the spectrogram of a WAV file

        Args:
            spectrogram (np.ndarray): Spectrogram of the WAV file.
            title (str, optional): Title of the plot. Defaults to None.
            transpose (bool, optional): Transpose the spectrogram. Defaults to True.
            invert (bool, optional): Invert the spectrogram. Defaults to True.
        """
        if transpose:
            spectrogram = spectrogram.T
        
        if invert:
            spectrogram = spectrogram[::-1]

        plt.figure(figsize=(15, 5))
        plt.imshow(spectrogram, aspect='auto', origin='lower')
        plt.title(f'Spectrogram: {title}')
        plt.xlabel('Time')
        plt.ylabel('Frequency')
        plt.colorbar()
        plt.tight_layout()
        plt.show()

    def __call__(self, audio_path: str, label: typing.Any):
        """
        Extract the spectrogram and label of a WAV file.

        Args:
            audio_path (str): Path to the WAV file.
            label (typing.Any): Label of the WAV file.

        Returns:
            Tuple[np.ndarray, typing.Any]: Spectrogram of the WAV file and its label.
        """
        return self.get_spectrogram(audio_path, self.frame_length, self.frame_step, self.fft_length), label


import os
import yaml

class BaseModelConfigs:
    def __init__(self):
        self.model_path = None

    def serialize(self):
        # get object attributes
        return self.__dict__

    def save(self, name: str='configs.yaml'):
        if self.model_path is None:
            raise Exception("Model path is not specified")

        # create directory if not exist
        if not os.path.exists(self.model_path):
            os.makedirs(self.model_path)

        with open(os.path.join(self.model_path, name), 'w') as f:
            yaml.dump(self.serialize(), f)

    @staticmethod
    def load(configs_path: str):
        with open(configs_path, 'r') as f:
            configs = yaml.load(f, Loader=yaml.FullLoader)

        config = BaseModelConfigs()
        for key, value in configs.items():
            setattr(config, key, value)

        return config


import os
import logging
import numpy as np

class Callback:
    """ Base class used to build new callbacks."""
    def __init__(
        self, 
        monitor: str = "val_loss"
    ) -> None:
        self.monitor = monitor
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(logging.INFO)

    def on_train_begin(self, logs=None):
        pass

    def on_train_end(self, logs=None):
        pass

    def on_train_batch_begin(self, batch: int, logs=None):
        pass

    def on_train_batch_end(self, batch: int, logs=None):
        pass

    def on_test_begin(self, logs=None):
        pass

    def on_test_end(self, logs=None):
        pass

    def on_test_batch_begin(self, batch: int, logs=None):
        pass

    def on_test_batch_end(self, batch: int, logs=None):
        pass

    def on_epoch_begin(self, epoch: int, logs=None):
        pass

    def on_epoch_end(self, epoch: int, logs=None):
        pass

    def on_batch_begin(self, batch: int, logs=None):
        pass

    def on_batch_end(self, batch: int, logs=None):
        pass

    def get_monitor_value(self, logs: dict):
        logs = logs or {}
        monitor_value = logs.get(self.monitor)
        if monitor_value is None:
            logging.warning(
                "Early stopping conditioned on metric `%s` "
                "which is not available. Available metrics are: %s",
                self.monitor,
                ",".join(list(logs.keys())),
            )
        return monitor_value

class EarlyStopping(Callback):
    def __init__(
        self, 
        monitor: str = "val_loss",
        min_delta: float = 0.0, 
        patience: int = 0, 
        verbose: bool = False,
        mode: str = "max_equal",
        ):
        super(EarlyStopping, self).__init__()

        self.monitor = monitor
        self.min_delta = min_delta
        self.patience = patience
        self.verbose = verbose
        self.mode = mode
        self.wait = None
        self.stopped_epoch = None
        self.best = None

        if self.mode not in ["min", "max", "max_equal", "min_equal"]:
            raise ValueError(
                "EarlyStopping mode %s is unknown, "
                "please choose one of min, max, max_equal, min_equal" % self.mode
            )
        
    def on_train_begin(self, logs=None):
        self.wait = 0
        self.stopped_epoch = 0
        self.best = np.inf if self.mode == "min" or self.mode == "min_equal" else -np.Inf

    def on_epoch_end(self, epoch: int, logs=None):
        current = self.get_monitor_value(logs)
        if current is None:
            return

        if self.mode == "min" and np.less(current, self.best - self.min_delta):
            self.best = current
            self.wait = 0
        elif self.mode == "max" and np.greater(current, self.best + self.min_delta):
            self.best = current
            self.wait = 0
        elif self.mode == "min_equal" and np.less_equal(current, self.best - self.min_delta):
            self.best = current
            self.wait = 0
        elif self.mode == "max_equal" and np.greater_equal(current, self.best + self.min_delta):
            self.best = current
            self.wait = 0
        else:
            self.wait += 1
            if self.wait >= self.patience:
                self.stopped_epoch = epoch
                self.model.stop_training = True

    def on_train_end(self, logs=None):
        if self.stopped_epoch > 0 and self.verbose:
            self.logger.info(f"Epoch {self.stopped_epoch}: early stopping")


class ModelCheckpoint(Callback):
    """ ModelCheckpoint callback to save the model after every epoch or the best model across all epochs."""
    def __init__(
        self, 
        filepath: str,
        monitor: str = "val_loss",
        verbose: bool = False,
        save_best_only: bool = True,
        mode: str = "min",
        ) -> None:
        """ ModelCheckpoint callback to save the model after every epoch or the best model across all epochs
        
        Args:
            filepath (str): path to save the model file
            monitor (str, optional): metric to monitor. Defaults to "val_loss".
            verbose (bool, optional): verbosity mode. Defaults to False.
            save_best_only (bool, optional): if True, the latest best model according to the quantity monitored will not be overwritten. Defaults to True.
            mode (str, optional): one of {min, max, max_equal, min_equal}. Defaults to "min".
        """
        super(ModelCheckpoint, self).__init__()

        self.filepath = filepath
        self.monitor = monitor
        self.verbose = verbose
        self.mode = mode
        self.save_best_only = save_best_only
        self.best = None

        if self.mode not in ["min", "max", "max_equal", "min_equal"]:
            raise ValueError(
                "ModelCheckpoint mode %s is unknown, "
                "please choose one of min, max, max_equal, min_equal" % self.mode
            )
        
        if self.mode == "min": self.monitor_op = np.less
        elif self.mode == "max": self.monitor_op = np.greater
        elif self.mode == "min_equal": self.monitor_op = np.less_equal
        elif self.mode == "max_equal": self.monitor_op = np.greater_equal
        
    def on_train_begin(self, logs=None):
        self.best = np.inf if self.mode == "min" or self.mode == "min_equal" else -np.Inf

        # create directory if not exist
        if not os.path.exists(os.path.dirname(self.filepath)):
            os.makedirs(os.path.dirname(self.filepath))

    def on_epoch_end(self, epoch: int, logs=None):
        current = self.get_monitor_value(logs)
        if current is None:
            return

        if self.monitor_op(current, self.best):
            previous = self.best
            self.best = current
            self.save_model(epoch, current, previous)
        else:
            if not self.save_best_only:
                self.save_model(epoch, current, previous=None)

    def save_model(self, epoch: int, best: float, previous: float = None):
        """ Save model to filepath
        
        Args:
            epoch (int): current epoch
            best (float): current best value
            previous (float, optional): previous best value. Defaults to None.
        """
        if self.verbose:
            if previous is None:
                self.logger.info(f"Epoch {epoch}: {self.monitor} got {best:.5f}, saving model to {self.filepath}")
            else:
                self.logger.info(f"Epoch {epoch}: {self.monitor} improved from {previous:.5f} to {best:.5f}, saving model to {self.filepath}")

        self.model.save(self.filepath)


import unittest

from mltu.utils.text_utils import edit_distance, get_cer, get_wer 

class TestTextUtils(unittest.TestCase):

    def test_edit_distance(self):
        """ This unit test includes several test cases to cover different scenarios, including no errors, 
        substitution errors, insertion errors, deletion errors, and a more complex case with multiple 
        errors. It also includes a test case for empty input.
        """
        # Test simple case with no errors
        prediction_tokens = ['A', 'B', 'C']
        reference_tokens = ['A', 'B', 'C']
        self.assertEqual(edit_distance(prediction_tokens, reference_tokens), 0)
        
        # Test simple case with one substitution error
        prediction_tokens = ['A', 'B', 'D']
        reference_tokens = ['A', 'B', 'C']
        self.assertEqual(edit_distance(prediction_tokens, reference_tokens), 1)
        
        # Test simple case with one insertion error
        prediction_tokens = ['A', 'B', 'C']
        reference_tokens = ['A', 'B', 'C', 'D']
        self.assertEqual(edit_distance(prediction_tokens, reference_tokens), 1)
        
        # Test simple case with one deletion error
        prediction_tokens = ['A', 'B']
        reference_tokens = ['A', 'B', 'C']
        self.assertEqual(edit_distance(prediction_tokens, reference_tokens), 1)
        
        # Test more complex case with multiple errors
        prediction_tokens = ['A', 'B', 'C', 'D', 'E']
        reference_tokens = ['A', 'C', 'B', 'F', 'E']
        self.assertEqual(edit_distance(prediction_tokens, reference_tokens), 3)
        
        # Test empty input
        prediction_tokens = []
        reference_tokens = []
        self.assertEqual(edit_distance(prediction_tokens, reference_tokens), 0)

    def test_get_cer(self):
        # Test simple case with no errors
        preds = ['A B C']
        target = ['A B C']
        self.assertEqual(get_cer(preds, target), 0)
        
        # Test simple case with one character error
        preds = ['A B C']
        target = ['A B D']
        self.assertEqual(get_cer(preds, target), 1/5)
        
        # Test simple case with multiple character errors
        preds = ['A B C']
        target = ['D E F']
        self.assertEqual(get_cer(preds, target), 3/5)
        
        # Test empty input
        preds = []
        target = []
        self.assertEqual(get_cer(preds, target), 0)

        # Test simple case with different word lengths
        preds = ['ABC']
        target = ['ABCDEFG']
        self.assertEqual(get_cer(preds, target), 4/7)

    def test_get_wer(self):
        # Test simple case with no errors
        preds = 'A B C'
        target = 'A B C'
        self.assertEqual(get_wer(preds, target), 0)
        
        # Test simple case with one word error
        preds = 'A B C'
        target = 'A B D'
        self.assertEqual(get_wer(preds, target), 1/3)
        
        # Test simple case with multiple word errors
        preds = 'A B C'
        target = 'D E F'
        self.assertEqual(get_wer(preds, target), 1)
        
        # Test empty input
        preds = ""
        target = ""
        self.assertEqual(get_wer(preds, target), 0)

        # Test simple case with different sentence lengths
        preds = ['ABC']
        target = ['ABC DEF']
        self.assertEqual(get_wer(preds, target), 1)

if __name__ == '__main__':
    unittest.main()